#include <sys/errno.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "spho/err.h"

#include "msph_token.h"


#define MSPH_ERR_SYS		0x0001
#define MSPH_ERR_TOOLONG	0x0002

#define MSPH_ERR(ctx, e)	SPHO_ERR(ctx, e)
#define MSPH_TOKS_ERR(toks, e)	MSPH_ERR((toks)->ctx, e)

struct msph_token_matcher {
	size_t		off;
	size_t		matchlen;

	const int	type;
} msph_matcher[] = {
	{ 0, 0, TOK_START },
	{ 0, 0, TOK_IDENT },
	{ 0, 0, TOK_END }
};

static ssize_t	src_file_fill_buf(struct msph_ctx *,
		    struct msph_token_src_file *);
static int	tok_match(struct msph_ctx *, struct msph_token_src *,
		    struct msph_token_matcher *);
static int	tok_commit(struct msph_ctx *, struct msph_token_src *, 
		    struct msph_token_matcher *, struct msph_token *);
static int	char_at(struct msph_ctx *, struct msph_token_src *, size_t,
		    char *);
static int	fromcbuf_charcpy(char *, const char *, size_t, size_t, size_t);
static int	file_char_at(struct msph_ctx *, struct msph_token_src *, size_t,
		    char *out);
static int	read_single_tok(struct msph_token *, struct msph_token_stream *);


struct msph_token_stream *
msph_token_stream_fopen(struct msph_ctx *ctx, const char *path)
{
	FILE *f;
	size_t res;
	struct msph_token_stream *ret;

	if ((f = fopen(path, "r")) == NULL) {
		MSPH_ERR(ctx, MSPH_ERR_SYS);
		return (NULL);
	}

	if ((ret = calloc(1, sizeof(struct msph_token_stream))) == NULL) {
		MSPH_ERR(ctx, MSPH_ERR_SYS);
		goto err;
	}

	ret->ctx = ctx;
	ret->src.type = MSPH_TOKEN_SRC_FILE;
	ret->src.inner.file.f = f;
	ret->src.inner.file.pos = 0;
	ret->src.inner.file.end = 0;

	res = strlcpy(ret->src.inner.file.name, path,
	    sizeof(ret->src.inner.file.name));
	if (res >= sizeof(ret->src.inner.file.name)) {
		MSPH_ERR(ctx, MSPH_ERR_TOOLONG);
		goto err;
	}

	return (ret);
err:
	if (fclose(f) == EOF)
		abort();

	if (ret != NULL)
		free(ret);

	return (NULL);
}

struct msph_token_stream *
msph_token_stream_frombuf(struct msph_ctx *ctx, const char *buf, size_t len)
{
	struct msph_token_stream *ret;

	if ((ret = calloc(1, sizeof(struct msph_token_stream))) == NULL) {
		MSPH_ERR(ctx, MSPH_ERR_SYS);
		return (NULL);
	}

	ret->ctx = ctx;
	ret->src.type = MSPH_TOKEN_SRC_STR;
	ret->src.inner.str.s = buf;
	ret->src.inner.str.len = len;
	ret->src.inner.str.pos = 0;

	return (ret);
}


int
msph_token_stream_close(struct msph_token_stream *s)
{
	int ret;

	ret = -1;

	switch (s->src.type) {
	case MSPH_TOKEN_SRC_FILE:
		ret = fclose(s->src.inner.file.f);
		break;
	case MSPH_TOKEN_SRC_STR:
		ret = 0;
		break;
	default:
		break;
	}

	return (ret);
}

/* -1 or num tokens read */
ssize_t
msph_token_stream_read_tok(struct msph_token *ptr, size_t n,
    struct msph_token_stream *s)
{
	ssize_t ret;
	size_t i;
	int res;

	ret = 0;
	res = -1;
	while ((res = read_single_tok(&ptr[i], s)) == 1) {
		if (res == -1) {
			ret = -1;
			break;
		}
		ret++;
	}

	return (ret);
}

/* 1: success, 0: failed match, -1: error */
static int
read_single_tok(struct msph_token *ptr, struct msph_token_stream *s)
{
	int res;
	size_t m;
	size_t max_m;
	struct msph_ctx *ctx;
	struct msph_token_src *src;

	ctx = s->ctx;
	src = &s->src;

	max_m = 0;
	for (m = 0; msph_matcher[m].type != TOK_END; m++) {
		res = tok_match(ctx, src, &msph_matcher[m]);

		if (res == -1)
			return (-1);

		if (res == 0 &&
		    msph_matcher[m].matchlen > msph_matcher[max_m].matchlen) {
			max_m = m;
		}
	}

	if (max_m == 0)
		return (0);

	if (tok_commit(ctx, src, &msph_matcher[max_m], ptr) == -1)
		return (-1);

	return (1);
}


#define BUF_LEN(b)	(sizeof(b) / sizeof((b)[0]))

static ssize_t
src_file_fill_buf(struct msph_ctx *ctx, struct msph_token_src_file *file)
{
	ssize_t ret;
	size_t nread, maxread;

	ret = nread = maxread = 0;
	do {
		if (file->end < file->pos)
			maxread = file->pos - file->end;
		else
			maxread = BUF_LEN(file->buf) - file->end;

		if (maxread == 0) {
			MSPH_ERR(ctx, MSPH_ERR_TOOLONG);
			return (-1);
		}

		nread = fread(&file->buf[file->end], sizeof(file->buf[0]),
		    maxread, file->f);

		ret += nread;
		file->end = (file->end + nread) % BUF_LEN(file->buf);

		if (nread < maxread) {
			if (ferror(file->f)) {
				MSPH_ERR(ctx, MSPH_ERR_SYS);
				return (-1);
			}
			break;
		}

	} while (file->end != file->pos);

	return (ret);
}

/* reads a single char from the circular buffer in src */
static int
file_char_at(struct msph_ctx *ctx, struct msph_token_src *src, size_t i,
    char *out)
{
	int ret;
	ssize_t fill;
	struct msph_token_src_file *file;

	SPHO_PRECOND(s != NULL);
	SPHO_PRECOND(s->src.type == MSPH_TOKEN_SRC_FILE);

	ret = 0;
	file = &src->inner.file;

	do {
		/* simplest case */
		if (file->pos + i < file->end) {
			*out = file->buf[file->pos + i];
			return (1);
		}
		/* wrap around */
		if (file->end < file->pos &&
		    ((file->pos + i) % BUF_LEN(file->buf)) < file->end) {
			*out = file->buf[(file->pos + i) % BUF_LEN(file->buf)];
			return (1);
		}

		if (feof(file->f))
			return (0);
		if (src_file_fill_buf(ctx, file) == -1)
			return (-1);
	} while (ret++);

	return (-1);
}

static int
char_at(struct msph_ctx *ctx, struct msph_token_src *src, size_t i, char *out)
{
	int ret;
	struct msph_token_src_str *str;

	switch (src->type) {
	case MSPH_TOKEN_SRC_FILE:
		return (file_char_at(ctx, src, i, out));
	case MSPH_TOKEN_SRC_STR:
		str = &src->inner.str;
		if (str->pos + i < str->len) {
			*out = str->s[str->pos + i];
			return (1);
		}
		return (0);
	default:
		break;
	}

	return (ret);
}

static int
fromcbuf_charcpy(char *dst, const char *src, size_t src_len, size_t src_pos,
    size_t ncpy)
{
	size_t cpy1, cpy2;
	if (src_len < ncpy) {
		return (-1);
	}

	cpy1 = src_pos + ncpy < src_len ? ncpy : src_len - src_pos;
	cpy2 = ncpy - cpy1;

	memcpy(dst, &src[src_pos], cpy1 * sizeof(src[0]));

	if (! cpy2)
		return (0);

	memcpy(dst, &src[0], cpy2 * sizeof(src[0]));

	return (0);
}

static int
tok_match(struct msph_ctx *ctx, struct msph_token_src *src,
    struct msph_token_matcher *m)
{
	int res;
	int more;
	char chr;
	const char *match_str;
	size_t off, len;
	SPHO_PRECOND(m != NULL && s != NULL);

	m->matchlen = 0;

#define MATCH_CHAR(c) \
	do { \
		if ((res = char_at(ctx, src, 0, &chr)) == -1) \
			return (-1); \
		else if (res == 0) \
			return (0); \
\
		if (chr == (c)) { \
			m->matchlen = 1; \
		} \
		return (0); \
	} while (0)

#define MATCH_STR(str) \
	do { \
		match_str = str; \
		len = strlen(match_str); \
		for (off = 0; off < len; off++) { \
			if ((res = char_at(ctx, src, off, &chr)) == -1) \
				return (-1); \
			else if (res == 0) \
				return (0); \
\
			if (chr != match_str[off]) \
				break; \
		} \
		if (off == len) \
			m->matchlen = len; \
		return (0); \
	} while (0)

	switch (m->type) {
	case TOK_LBRACE:
		MATCH_CHAR('{');
	case TOK_RBRACE:
		MATCH_CHAR('}');
	case TOK_LBRAK:
		MATCH_CHAR('[');
	case TOK_RBRAK:
		MATCH_CHAR(']');
	case TOK_LPAREN:
		MATCH_CHAR('(');
	case TOK_RPAREN:
		MATCH_CHAR(')');
	case TOK_COLON:
		MATCH_CHAR(':');
	case TOK_EQUALS:
		MATCH_CHAR('=');
	case TOK_AMP:
		MATCH_CHAR('&');
	case TOK_PIPE:
		MATCH_CHAR('|');
	case TOK_RARROW:
		MATCH_STR("=>");
	case TOK_SUB:
		MATCH_STR("<:");
	case TOK_KW_TYPE:
		MATCH_STR("type");
	case TOK_KW_NOMINAL:
		MATCH_STR("nominal");
	case TOK_KW_MEMBER:
		MATCH_STR("member");
	case TOK_KW_CHECK:
		MATCH_STR("check");
	case TOK_KW_BOX:
		MATCH_STR("box");
	case TOK_KW_FORALL:
		MATCH_STR("forall");
	case TOK_CONST_TRUE:
		MATCH_STR("True");
	case TOK_CONST_FALSE:
		MATCH_STR("False");
	case TOK_IDENT:
		off = 0;
		while ((res = char_at(ctx, src, off++, &chr)) == 1) {
			if (! isalnum(chr))
				break;
			m->matchlen++;
		}
		if (res == -1)
			return (-1);
		return (0);
	default:
		SPHO_ASSERT(0);
		return (-1);
		break;
	}
#undef MATCH_CHAR
#undef MATCH_STR
}

#define TOK_HAS_DATA(type)	(type == TOK_IDENT)
static int
tok_commit(struct msph_ctx *ctx, struct msph_token_src *src,
    struct msph_token_matcher *m, struct msph_token *ptr)
{
	size_t pos_old;
	struct msph_token_src_str *str;
	struct msph_token_src_file *file;

	SPHO_PRECOND(p != NULL && m != NULL);
	SPHO_PRECOND(m->matchlen != 0);

	switch (src->type) {
	case MSPH_TOKEN_SRC_FILE:
		file = &src->inner.file;
		pos_old = file->pos;

		file->pos += m->matchlen;
		file->pos %= BUF_LEN(file->buf);
		SPHO_ASSERT(file->pos < BUF_LEN(file->buf) ||
		    file->pos < pos_old);

		ptr->type = m->type;
		if (! TOK_HAS_DATA(ptr->type))
			return (0);

		if (m->matchlen >= sizeof(ptr->d.s.buf)) {
			MSPH_ERR(ctx, MSPH_ERR_TOOLONG);
			return (-1);
		}

		if (fromcbuf_charcpy(ptr->d.s.buf, file->buf, sizeof(file->buf),
			    pos_old, m->matchlen) == -1) {
			MSPH_ERR(ctx, MSPH_ERR_TOOLONG);
			return (-1);
		}

		ptr->d.s.buf[m->matchlen] = '\0';
		return (0);

	case MSPH_TOKEN_SRC_STR:
		str = &src->inner.str;
		pos_old = str->pos;

		str->pos += m->matchlen;
		SPHO_ASSERT(str->pos <= str->len);

		ptr->type = m->type;
		if (! TOK_HAS_DATA(ptr->type))
			return (0);
		
		if (m->matchlen >= sizeof(ptr->d.s.buf)) {
			MSPH_ERR(ctx, MSPH_ERR_TOOLONG);
			return (-1);
		}

		memcpy(ptr->d.s.buf, str->s, m->matchlen *
		    sizeof(str->s[0]));
		ptr->d.s.buf[m->matchlen] = '\0';

		return (0);
	default:
		return (-1);
	}
}






